import torch
from torch import nn

from .general import BoundModule
from .bounds import IntervalBounds, LinearBounds


class Cat(nn.Module):
    def __init__(self, subnetwork):
        super().__init__()

        self.subnetwork = subnetwork

    def forward(self, x):
        y = self.subnetwork(x)
        return torch.cat([x, y], dim=-1)


class BoundCat(BoundModule):
    def __init__(self, model, factory, **kwargs):
        super().__init__(model, factory)

        self.subnetwork = factory.build(model.subnetwork)
        self.in_size = None

    @property
    def need_relaxation(self):
        return self.subnetwork.need_relaxation

    def clear_relaxation(self):
        self.subnetwork.clear_relaxation()

    def backward_relaxation(self, region):
        assert self.subnetwork.need_relaxation

        return self.subnetwork.backward_relaxation(region)

    def crown_backward(self, linear_bounds, optimize):
        assert self.in_size is not None

        residual_linear_bounds = self.subnetwork.crown_backward(linear_bounds[..., self.in_size:], optimize)
        linear_bounds = linear_bounds[..., :self.in_size]

        if linear_bounds.lower is None:
            lower = None
        else:
            lower = (linear_bounds.lower[0] + residual_linear_bounds.lower[0], residual_linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = (linear_bounds.upper[0] + residual_linear_bounds.upper[0], residual_linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False):
        residual_bounds = self.subnetwork.ibp_forward(bounds, save_relaxation=save_relaxation)

        # Order matters here! - must match forward on Cat class
        lower = torch.cat([bounds.lower, residual_bounds.lower], dim=-1)
        upper = torch.cat([bounds.upper, residual_bounds.upper], dim=-1)
        return IntervalBounds(bounds.region, lower, upper)

    def propagate_size(self, in_size):
        self.in_size = in_size
        out_size = self.subnetwork.propagate_size(in_size)

        return in_size + out_size

    def bound_parameters(self):
        yield from self.subnetwork.bound_parameters()

    def reset_params(self):
        self.subnetwork.reset_params()

    def clip_params(self):
        self.subnetwork.clip_params()



from .activation import BoundActivation, BoundTanh, BoundReLU, BoundSigmoid, bisection
from .bounds import LinearBounds, IntervalBounds, HyperRectangle
from .factory import BoundModelFactory
from .general import BoundModule
from .linear import BoundLinear
from .sequential import BoundSequential
from .residual import Residual, BoundResidual
from .cat import Cat, BoundCat
from .parallel import Parallel, BoundParallel
from .bivariate import Add, Sub, BoundAdd, BoundSub
from .saturation import Clamp, BoundClamp



import torch
from torch import nn

from .bounds import IntervalBounds, LinearBounds
from .general import BoundModule


class Parallel(nn.Module):
    def __init__(self, subnetwork1, subnetwork2, split_size=None):
        super().__init__()

        self.split_size = split_size
        self.subnetwork1 = subnetwork1
        self.subnetwork2 = subnetwork2

    def forward(self, x):
        if self.split_size:
            x1, x2 = x[..., :self.split_size], x[..., self.split_size:]
        else:
            x1, x2 = x, x

        return torch.cat([self.subnetwork1(x1), self.subnetwork2(x2)], dim=-1)


class BoundParallel(BoundModule):
    def __init__(self, model, factory, **kwargs):
        super().__init__(model, factory)

        self.out_size1 = None
        self.subnetwork1 = factory.build(model.subnetwork1)
        self.subnetwork2 = factory.build(model.subnetwork2)

    @property
    def need_relaxation(self):
        return self.subnetwork1.need_relaxation or self.subnetwork2.need_relaxation

    def clear_relaxation(self):
        self.subnetwork1.clear_relaxation()
        self.subnetwork2.clear_relaxation()

    def backward_relaxation(self, region):
        if self.subnetwork1.need_relaxation:
            linear_bounds, relaxation_module = self.subnetwork1.backward_relaxation(region)
            return self.padding(linear_bounds, order=0), relaxation_module
        else:
            assert self.subnetwork2.need_relaxation
            linear_bounds, relaxation_module = self.subnetwork2.backward_relaxation(region)
            return self.padding(linear_bounds, order=1), relaxation_module

    def padding(self, linear_bounds, order=0):
        if self.module.split_size is None:
            return linear_bounds

        if order == 0:
            lowerA = torch.cat([linear_bounds.lower[0], torch.zeros_like(linear_bounds.lower[0])], dim=-1)
            upperA = torch.cat([linear_bounds.upper[0], torch.zeros_like(linear_bounds.upper[0])], dim=-1)
        else:
            lowerA = torch.cat([torch.zeros_like(linear_bounds.lower[0]), linear_bounds.lower[0]], dim=-1)
            upperA = torch.cat([torch.zeros_like(linear_bounds.upper[0]), linear_bounds.upper[0]], dim=-1)

        return LinearBounds(
            linear_bounds.region,
            (lowerA, linear_bounds.lower[1]),
            (upperA, linear_bounds.upper[1])
        )

    def crown_backward(self, linear_bounds, optimize):
        assert self.out_size1 is not None

        residual_linear_bounds1 = self.subnetwork1.crown_backward(linear_bounds[..., :self.out_size1], optimize)
        residual_linear_bounds2 = self.subnetwork2.crown_backward(linear_bounds[..., self.out_size1:], optimize)

        if linear_bounds.lower is None:
            lower = None
        else:
            if self.module.split_size is not None:
                lowerA = torch.cat([residual_linear_bounds1.lower[0], residual_linear_bounds2.lower[0]], dim=-1)
            else:
                lowerA = residual_linear_bounds1.lower[0] + residual_linear_bounds2.lower[0]

            lower = (lowerA, residual_linear_bounds1.lower[1] + residual_linear_bounds2.lower[1] - linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            if self.module.split_size is not None:
                upperA = torch.cat([residual_linear_bounds1.upper[0], residual_linear_bounds2.upper[0]], dim=-1)
            else:
                upperA = residual_linear_bounds1.upper[0] + residual_linear_bounds2.upper[0]

            upper = (upperA, residual_linear_bounds1.upper[1] + residual_linear_bounds2.upper[1] - linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False):
        if self.module.split_size is not None:
            bounds1, bounds2 = bounds[..., :self.module.split_size], bounds[..., self.module.split_size:]
        else:
            bounds1, bounds2 = bounds, bounds

        residual_bounds1 = self.subnetwork1.ibp_forward(bounds1, save_relaxation=save_relaxation)
        residual_bounds2 = self.subnetwork2.ibp_forward(bounds2, save_relaxation=save_relaxation)

        # Order matters here! - must match forward on Parallel class
        lower = torch.cat([residual_bounds1.lower, residual_bounds2.lower], dim=-1)
        upper = torch.cat([residual_bounds1.upper, residual_bounds2.upper], dim=-1)
        return IntervalBounds(bounds.region, lower, upper)

    def propagate_size(self, in_size):
        if self.module.split_size is not None:
            out_size1 = self.subnetwork1.propagate_size(self.module.split_size)
            out_size2 = self.subnetwork2.propagate_size(in_size - self.module.split_size)
        else:
            out_size1 = self.subnetwork1.propagate_size(in_size)
            out_size2 = self.subnetwork2.propagate_size(in_size)

        self.out_size1 = out_size1

        return out_size1 + out_size2

    def bound_parameters(self):
        yield from self.subnetwork1.bound_parameters()
        yield from self.subnetwork2.bound_parameters()

    def reset_params(self):
        self.subnetwork1.reset_params()
        self.subnetwork2.reset_params()

    def clip_params(self):
        self.subnetwork1.clip_params()
        self.subnetwork2.clip_params()



import torch
from torch import nn

from .general import BoundModule
from .bounds import IntervalBounds, LinearBounds


class Add(nn.Module):
    def __init__(self, network1, network2):
        super().__init__()

        self.network1, self.network2 = network1, network2

    def forward(self, x):
        return self.network1(x) + self.network2(x)


class BoundAdd(BoundModule):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.bound_network1 = factory.build(module.network1)
        self.bound_network2 = factory.build(module.network2)

    @property
    def need_relaxation(self):
        return self.bound_network1.need_relaxation or self.bound_network2.need_relaxation

    def clear_relaxation(self):
        self.bound_network1.clear_relaxation()
        self.bound_network2.clear_relaxation()

    def backward_relaxation(self, region):
        if self.bound_network1.need_relaxation:
            return self.bound_network1.backward_relaxation(region)
        else:
            assert self.bound_network2.need_relaxation
            return self.bound_network2.backward_relaxation(region)

    def crown_backward(self, linear_bounds, optimize):
        linear_bounds1 = self.bound_network1.crown_backward(linear_bounds, optimize)
        linear_bounds2 = self.bound_network2.crown_backward(linear_bounds, optimize)

        if linear_bounds.lower is None:
            lower = None
        else:
            lower = (linear_bounds1.lower[0] + linear_bounds2.lower[0], linear_bounds1.lower[1] + linear_bounds2.lower[1] - linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = (linear_bounds1.upper[0] + linear_bounds2.upper[0], linear_bounds1.upper[1] + linear_bounds2.upper[1] - linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False):
        bounds1 = self.bound_network1.ibp_forward(bounds, save_relaxation=save_relaxation)
        bounds2 = self.bound_network2.ibp_forward(bounds, save_relaxation=save_relaxation)

        return IntervalBounds(
            bounds.region,
            bounds1.lower + bounds2.lower,
            bounds1.upper + bounds2.upper
        )

    def propagate_size(self, in_size):
        out_size1 = self.bound_network1.propagate_size(in_size)
        out_size2 = self.bound_network2.propagate_size(in_size)

        assert out_size1 == out_size2

        return out_size1

    def bound_parameters(self):
        yield from self.bound_network1.bound_parameters()
        yield from self.bound_network2.bound_parameters()

    def reset_params(self):
        self.bound_network1.reset_params()
        self.bound_network2.reset_params()

    def clip_params(self):
        self.bound_network1.clip_params()
        self.bound_network2.clip_params()


class VectorAdd(nn.Module):
    def forward(self, x):
        assert x.size(-1) % 2 == 0
        half_size = x.size(-1) // 2

        return x[..., :half_size] + x[..., half_size:]


class BoundVectorAdd(BoundModule):
    @property
    def need_relaxation(self):
        return False

    def crown_backward(self, linear_bounds, optimize):
        if linear_bounds.lower is None:
            lower = None
        else:
            lower = (torch.cat(2 * [linear_bounds.lower[0]], dim=-1), linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = (torch.cat(2 * [linear_bounds.upper[0]], dim=-1), linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False):
        half_size = bounds.lower.size(-1) // 2

        return IntervalBounds(
            bounds.region,
            bounds.lower[..., :half_size] + bounds.lower[..., half_size:],
            bounds.upper[..., :half_size] + bounds.upper[..., half_size:]
        )

    def propagate_size(self, in_size):
        assert in_size % 2 == 0

        return in_size // 2


class Sub(nn.Module):
    def __init__(self, network1, network2):
        super().__init__()

        self.network1, self.network2 = network1, network2

    def forward(self, x):
        return self.network1(x) - self.network2(x)


class BoundSub(BoundModule):
    def __init__(self, module, factory, **kwargs):
        super().__init__(module, factory, **kwargs)

        self.bound_network1 = factory.build(module.network1)
        self.bound_network2 = factory.build(module.network2)

    @property
    def need_relaxation(self):
        return self.bound_network1.need_relaxation or self.bound_network2.need_relaxation

    def clear_relaxation(self):
        self.bound_network1.clear_relaxation()
        self.bound_network2.clear_relaxation()

    def backward_relaxation(self, region):
        if self.bound_network1.need_relaxation:
            return self.bound_network1.backward_relaxation(region)
        else:
            assert self.bound_network2.need_relaxation
            return self.bound_network2.backward_relaxation(region)

    def crown_backward(self, linear_bounds, optimize):
        input_bounds = LinearBounds(
            linear_bounds.region,
            (linear_bounds.lower[0], torch.zeros_like(linear_bounds.lower[1])) if linear_bounds.lower is not None else None,
            (linear_bounds.upper[0], torch.zeros_like(linear_bounds.upper[1])) if linear_bounds.upper is not None else None,
        )
        linear_bounds1 = self.bound_network1.crown_backward(input_bounds, optimize)

        input_bounds = LinearBounds(
            linear_bounds.region,
            (-linear_bounds.lower[0], torch.zeros_like(linear_bounds.lower[1])) if linear_bounds.lower is not None else None,
            (-linear_bounds.upper[0], torch.zeros_like(linear_bounds.upper[1])) if linear_bounds.upper is not None else None,
        )
        linear_bounds2 = self.bound_network2.crown_backward(input_bounds, optimize)

        if linear_bounds.lower is None:
            lower = None
        else:
            lower = (linear_bounds1.lower[0] + linear_bounds2.lower[0], linear_bounds1.lower[1] + linear_bounds2.lower[1] + linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = (linear_bounds1.upper[0] + linear_bounds2.upper[0], linear_bounds1.upper[1] + linear_bounds2.upper[1] + linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False):
        bounds1 = self.bound_network1.ibp_forward(bounds, save_relaxation=save_relaxation)
        bounds2 = self.bound_network2.ibp_forward(bounds, save_relaxation=save_relaxation)

        return IntervalBounds(
            bounds.region,
            bounds1.lower - bounds2.upper,
            bounds1.upper - bounds2.lower
        )

    def propagate_size(self, in_size):
        out_size1 = self.bound_network1.propagate_size(in_size)
        out_size2 = self.bound_network2.propagate_size(in_size)

        assert out_size1 == out_size2

        return out_size1

    def bound_parameters(self):
        yield from self.bound_network1.bound_parameters()
        yield from self.bound_network2.bound_parameters()

    def reset_params(self):
        self.bound_network1.reset_params()
        self.bound_network2.reset_params()

    def clip_params(self):
        self.bound_network1.clip_params()
        self.bound_network2.clip_params()


class VectorSub(nn.Module):
    def forward(self, x):
        assert x.size(-1) % 2 == 0
        half_size = x.size(-1) // 2

        return x[..., :half_size] - x[..., half_size:]


class BoundVectorSub(BoundModule):
    @property
    def need_relaxation(self):
        return False

    def crown_backward(self, linear_bounds, optimize):
        if linear_bounds.lower is None:
            lower = None
        else:
            lower = (torch.cat([linear_bounds.lower[0], -linear_bounds.lower[0]], dim=-1), linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = (torch.cat([linear_bounds.upper[0], -linear_bounds.upper[0]], dim=-1), linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    def ibp_forward(self, bounds, save_relaxation=False):
        half_size = bounds.lower.size(-1) // 2

        return IntervalBounds(
            bounds.region,
            bounds.lower[..., :half_size] - bounds.upper[..., half_size:],
            bounds.upper[..., :half_size] - bounds.lower[..., half_size:]
        )

    def propagate_size(self, in_size):
        assert in_size % 2 == 0

        return in_size // 2



from typing import Optional, Tuple

import torch

from .activation import assert_bound_order
from .general import BoundModule
from .bounds import LinearBounds, IntervalBounds


@torch.jit.script
def crown_backward_linear_jit(weight: torch.Tensor, bias: Optional[torch.Tensor], W_tilde: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    if bias is None:
        bias_acc = torch.tensor(0.0, device=W_tilde.device)
    else:
        bias = bias.unsqueeze(-2)
        bias_acc = bias.matmul(W_tilde.transpose(-1, -2)).squeeze(-2)

    if weight.dim() == 2:
        W_tilde = W_tilde.matmul(weight)
    else:
        if W_tilde.dim() == 3:
            W_tilde = W_tilde.unsqueeze(0)

        W_tilde = W_tilde.matmul(weight.unsqueeze(1))

    return W_tilde, bias_acc


@torch.jit.script
def ibp_forward_linear_jit(weight: torch.Tensor, bias: Optional[torch.Tensor], center: torch.Tensor, diff: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    center, diff = center.unsqueeze(-2), diff.unsqueeze(-2)

    weight = weight.transpose(-1, -2)

    w_mid = center.matmul(weight) + (bias.unsqueeze(-2) if bias is not None else torch.tensor(0.0, device=weight.device))
    w_diff = diff.matmul(weight.abs())

    lower = w_mid - w_diff
    lower = lower.squeeze(-2)

    upper = w_mid + w_diff
    upper = upper.squeeze(-2)

    return lower, upper


class BoundLinear(BoundModule):

    @property
    def need_relaxation(self):
        return False

    def crown_backward(self, linear_bounds, optimize):
        if linear_bounds.lower is None:
            lower = None
        else:
            lower = crown_backward_linear_jit(self.module.weight, self.module.bias, linear_bounds.lower[0])
            lower = (lower[0], lower[1] + linear_bounds.lower[1])

        if linear_bounds.upper is None:
            upper = None
        else:
            upper = crown_backward_linear_jit(self.module.weight, self.module.bias, linear_bounds.upper[0])
            upper = (upper[0], upper[1] + linear_bounds.upper[1])

        return LinearBounds(linear_bounds.region, lower, upper)

    @assert_bound_order
    def ibp_forward(self, bounds, save_relaxation=False):
        center, diff = bounds.center, bounds.width / 2
        lower, upper = ibp_forward_linear_jit(self.module.weight, self.module.bias, center, diff)

        return IntervalBounds(bounds.region, lower, upper)

    def propagate_size(self, in_size):
        return self.module.weight.size(-2)



import abc

import torch
from torch import nn
from torch.optim import Adam
from torch.optim.lr_scheduler import ExponentialLR

from .bounds import LinearBounds, IntervalBounds


class BoundModule(nn.Module, abc.ABC):
    def __init__(self, module, factory, **kwargs):
        super().__init__()
        self.module = module

        self.optimizer = kwargs.get('bounds_optimizer', Adam)
        self.bounds_iterations = kwargs.get('bounds_iterations', 40)
        self.lr = kwargs.get('bounds_lr', 1e-2)
        self.lr_decay = kwargs.get('bounds_lr_decay', 0.97)

    @torch.no_grad()
    def crown_relax(self, region):
        # No grad for relaxations improves accuracy and stabilizes training for CROWN.
        while self.need_relaxation:
            linear_bounds, module = self.backward_relaxation(region)
            module.set_relaxation(linear_bounds)

    @torch.no_grad()
    def ibp_relax(self, region):
        # No grad for relaxations improves accuracy. CROWN-IBP is already stable in training.
        bounds = IntervalBounds(region, region.lower, region.upper)
        self.ibp_forward(bounds, save_relaxation=True)

    def crown(self, region, bound_lower=True, bound_upper=True, alpha=False):
        return self.crown_with_relaxation(self.crown_relax, region, bound_lower, bound_upper, alpha)

    def crown_ibp(self, region, bound_lower=True, bound_upper=True, alpha=False):
        return self.crown_with_relaxation(self.ibp_relax, region, bound_lower, bound_upper, alpha)

    def crown_with_relaxation(self, relax, region, bound_lower, bound_upper, alpha):
        out_size = self.propagate_size(region.lower.size(-1))

        relax(region)

        if alpha:
            linear_bounds = self.alpha_crown(region, out_size, bound_lower, bound_upper)
        else:
            linear_bounds = self.initial_linear_bounds(region, out_size, lower=bound_lower, upper=bound_upper)
            linear_bounds = self.crown_backward(linear_bounds, False)

        self.clear_relaxation()
        return linear_bounds

    def initial_linear_bounds(self, region, out_size, lower=True, upper=True):
        W_tilde = torch.eye(out_size, device=region.lower.device)\
            .unsqueeze(-3).expand(*region.lower.size()[:-1], out_size, out_size)
        bias = torch.zeros((out_size,), device=region.lower.device)\
            .unsqueeze(-2).expand(*region.lower.size()[:-1], out_size)

        lower = (W_tilde, bias) if lower else None
        upper = (W_tilde, bias) if upper else None

        linear_bounds = LinearBounds(region, lower, upper)
        return linear_bounds

    @torch.enable_grad()
    def alpha_crown(self, region, out_size, bound_lower, bound_upper):
        params = list(self.bound_parameters())

        if bound_lower:
            lower = self.optimize_bounds(region, out_size, True, params).lower
        else:
            lower = None

        if bound_upper:
            upper = self.optimize_bounds(region, out_size, False, params).upper
        else:
            upper = None

        return LinearBounds(region, lower, upper)

    def optimize_bounds(self, region, out_size, lower, params):
        optimizer = self.optimizer(params, self.lr)
        scheduler = ExponentialLR(optimizer, self.lr_decay)

        self.reset_params()

        for iteration in range(self.bounds_iterations):
            linear_bounds = self.initial_linear_bounds(region, out_size, lower=lower, upper=not lower)
            linear_bounds = self.crown_backward(linear_bounds, True)
            interval_bounds = linear_bounds.concretize()

            loss = (-interval_bounds.lower if lower else interval_bounds.upper).sum()

            optimizer.zero_grad(set_to_none=True)
            loss.backward()
            optimizer.step()
            scheduler.step()

            self.clip_params()  # Projected Gradient Descent

        with torch.no_grad():
            linear_bounds = self.initial_linear_bounds(region, out_size, lower=lower, upper=not lower)
            return self.crown_backward(linear_bounds, True)

    @property
    @abc.abstractmethod
    def need_relaxation(self):
        raise NotImplementedError()

    def clear_relaxation(self):
        pass

    def set_relaxation(self, linear_bounds):
        pass

    def backward_relaxation(self, region):
        pass

    @abc.abstractmethod
    def crown_backward(self, linear_bounds, optimize):
        raise NotImplementedError()

    def ibp(self, region):
        bounds = IntervalBounds(region, region.lower, region.upper)
        return self.ibp_forward(bounds)

    @abc.abstractmethod
    def ibp_forward(self, bounds, save_relaxation=False):
        raise NotImplementedError()

    @abc.abstractmethod
    def propagate_size(self, in_size):
        raise NotImplementedError()

    def forward(self, *args, **kwargs):
        return self.module(*args, **kwargs)

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        return self.module.state_dict(destination=destination, prefix=prefix, keep_vars=keep_vars)

    def load_state_dict(self, state_dict: 'OrderedDict[str, Tensor]', strict: bool = True):
        # Assume that submodules are constructed with the factory, since that produces
        # references down the bound module graph corresponding to the model graph.
        # Then loading the state dict at the top level automatically populates the weights in all sub-bound modules.
        self.module.load_state_dict(state_dict=state_dict, strict=strict)

    def bound_parameters(self):
        return []

    def reset_params(self):
        pass

    def clip_params(self):
        pass
